import sys
from typing import ClassVar, Literal
from pydantic import BaseModel
import sympy
from sympy import (
    Ge,
    Gt,
    Le,
    Lt,
    Ne,
    Not,
    Symbol,
    Eq,
    Or,
    And,
    simplify,
    simplify_logic,
    SOPform,
)
from sympy.logic.boolalg import Boolean


class SympyToSqlHelper:
    columns: list[str]

    @staticmethod
    def to_sql_where(expr: Boolean) -> str:
        # expr = simplify_logic(expr=expr, form="dnf")
        return SympyToSqlHelper._to_sql_impl(expr)

    @staticmethod
    def _to_sql_impl(expr: Boolean) -> str:
        """
        Recursively translates a Sympy expression into a sql where clause.
        """
        # --- Logical Operators ---
        conv = [SympyToSqlHelper._to_sql_impl(arg) for arg in expr.args]

        if isinstance(expr, And):
            return f"( {' AND '.join(conv)} )"

        if isinstance(expr, Or):
            return f"( {' OR '.join(conv)} )"

        if isinstance(expr, Not):
            return f" NOT {conv[0]} "

        # --- Comparison Operators ---
        if isinstance(expr, Eq):
            return f" {conv[0]}={conv[1]} "
        if isinstance(expr, Ne):
            return f" {conv[0]}<>{conv[1]} "
        if isinstance(expr, Ge):
            return f" {conv[0]}>={conv[1]} "
        if isinstance(expr, Gt):
            return f" {conv[0]}>{conv[1]} "
        if isinstance(expr, Lt):
            return f" {conv[0]}<{conv[1]} "
        if isinstance(expr, Le):
            return f" {conv[0]}<={conv[1]} "

        # --- Base Cases (End of Recursion) ---
        if isinstance(expr, Symbol):
            name = str(expr)
            if name in SympyToSqlHelper.columns:
                return f'"{name}"'
            else:
                return f"'{name}'"

        if isinstance(expr, sympy.core.numbers.Number):
            return f"{float(expr)}"

        raise TypeError(f"Unknown Sympy type to convert: {type(expr)}")


class SympyContainer(BaseModel):
    feat_name: str
    operator: Literal["=", "<", "<=", ">", ">=", "!=", "not in", "in", "between"]
    value: str | int | float | list[str] | list[int] | list[float]
    assumptions: dict[str, bool]
    feature_symbol_map: ClassVar[dict[str, Symbol]] = {}
    feature_value_symbol_map: ClassVar[dict[str, dict[int | str, Symbol]]] = {}

    @staticmethod
    def get_value_symbol(feature_name: str, value) -> Symbol | int:
        if isinstance(value, int):
            return value
        return SympyContainer.feature_value_symbol_map.setdefault(
            feature_name, {}
        ).setdefault(value, Symbol(value))

    @staticmethod
    def get_symbol(
        feature_name: str, assumptions: dict[str, bool] | None = None
    ) -> Symbol:
        assumptions = assumptions if assumptions is not None else {}
        return SympyContainer.feature_symbol_map.setdefault(
            feature_name, Symbol(feature_name, **assumptions)
        )

    def get_sympy(self) -> Boolean:
        x = self.get_symbol(self.feat_name, self.assumptions)

        match self.operator:
            case "=":
                return Eq(x, self.get_value_symbol(self.feat_name, self.value))
            case "!=":
                return Ne(x, self.get_value_symbol(self.feat_name, self.value))
            case "in":
                return Or(
                    *[
                        Eq(x, self.get_value_symbol(self.feat_name, value))
                        for value in self.value
                    ]
                )
            case "not in":
                return Not(
                    Or(
                        *[
                            Eq(x, self.get_value_symbol(self.feat_name, value))
                            for value in self.value
                        ]
                    )
                )
            case "<":
                return Lt(x, self.value)
            case "<=":
                return Le(x, self.value)
            case ">":
                return Gt(x, self.value)
            case ">=":
                return Ge(x, self.value)
            case "between":
                return And(Gt(x, self.value[0]), Le(x, self.value[1]))
        raise NotImplementedError(f"{self.operator} not implemented.")


class Filter:
    def __init__(self, combine: Boolean):
        self.combine = combine

    def sql(self, do_invert: bool = False) -> str:
        combine = Not(self.combine) if do_invert else self.combine
        return SympyToSqlHelper.to_sql_where(combine)



import itertools
import sys
from typing import Optional
from pydantic import BaseModel
from pilz.model.datacard import DataCard, Feature, FeatureType
from pilz.model.dataframes import (
    CatCats,
    CombinedCategorizedFeature,
    ContCats,
    TrainDataframes,
    CategorizedFeature,
)
from pilz.model.filter import (
    Filter,
    SympyContainer,
    SympyToSqlHelper,
)
from pilz.model.settings import TrainSettings
from pilz.model.pilz import Pilz, Spore
import polars as pl


import logging

from pilz.service.darkwing import Darkwing

logger = logging.getLogger(__name__)


class Train(BaseModel):
    dc: DataCard
    settings: TrainSettings
    darkwing: Optional[Darkwing] = None

    def model_post_init(self, __context) -> None:
        self.darkwing = Darkwing(dc=self.dc)
        super().model_post_init(__context)

    def run(self):
        logger.info("start training")
        SympyToSqlHelper.columns = self.dc.feature_names
        for n in range(self.settings.n):
            for cat in self.dc.target.values:
                if self.settings.pilz_exists(target=cat, n=n):
                    logger.info(f"tree {n} for label {cat} already exists")
                    continue

                logger.info(f"train tree {n} for label {cat}")

                train_res = Pilz(
                    spore=self.train_pilz(
                        target_filter=self.gen_target_filter(cat=cat),
                        path_filter=[],
                    ),
                    target=cat,
                )
                self.settings.save_pilz(pilz=train_res, n=n)

    def gen_target_filter(self, cat) -> Filter:
        return Filter(
            combine=SympyContainer(
                feat_name=self.dc.target.feature_name,
                operator="=",
                value=cat,
                assumptions={},
            ).get_sympy()
        )

    def make_spore(
        self, path_filter: list[Filter], depth: str, train_df: TrainDataframes
    ) -> Spore:
        score = train_df.score()
        logging.info(
            f"final spore depth: {depth}, {score} non target {train_df.non_target_df_size} target: {train_df.target_df_size}"
        )
        return [
            Spore(
                cut=[fil.sql() for fil in path_filter],
                score=score,
                depth=depth,
            )
        ]

    def train_pilz(
        self, target_filter: Filter, path_filter: list[Filter], depth: str = ""
    ) -> list[Spore]:
        train_df = self.darkwing.read_akt_train(
            targer_filter=target_filter,
            train_settings=self.settings,
            akt_filters=path_filter,
        )

        if train_df.is_final_size() or len(depth) >= self.settings.max_depth:
            return self.make_spore(
                path_filter=path_filter, depth=depth, train_df=train_df
            )

        self.cater(train_df=train_df)
        left_filter, neutral_filter, right_filter = self.counter(train_df=train_df)
        if left_filter is None and right_filter is None:
            # neutral must then be true.
            logger.info("Best cut lands all in neutral")
            return self.make_spore(
                path_filter=path_filter, depth=depth, train_df=train_df
            )

        left_spores = (
            self.train_pilz(
                target_filter=target_filter,
                path_filter=path_filter + [left_filter],
                depth=depth + "l",
            )
            if left_filter is not None
            else []
        )
        neutral_spores = (
            self.train_pilz(
                target_filter=target_filter,
                path_filter=path_filter + [neutral_filter],
                depth=depth + "n",
            )
            if neutral_filter is not None
            else []
        )

        right_spores = (
            self.train_pilz(
                target_filter=target_filter,
                path_filter=path_filter + [right_filter],
                depth=depth + "r",
            )
            if right_filter is not None
            else []
        )
        return left_spores + neutral_spores + right_spores

    def counter(
        self, train_df: TrainDataframes
    ) -> tuple[Filter, Filter | None, Filter]:
        sorted_train_feats = sorted(
            train_df.train_features,
            key=lambda x: x.calc_diff(),
            reverse=True,
        )
        best_feat = sorted_train_feats[0]
        best_feat_diff = best_feat.calc_diff()

        for dim in range(2, self.settings.n_dims + 1):
            counter = 0
            for comb in itertools.combinations(sorted_train_feats, r=dim):
                counter += 1
                akt_feature = CombinedCategorizedFeature(
                    comb,
                    non_target_size=train_df.n_count_non_target,
                    target_size=train_df.n_count_target,
                    neutral_faktor=self.settings.neutral_faktor,
                )
                if akt_feature.calc_diff() > best_feat_diff:
                    best_feat = akt_feature
                    best_feat_diff = akt_feature.calc_diff()

                if (
                    self.settings.calcs_per_dim
                    and counter > self.settings.calcs_per_dim
                ):
                    break

        return best_feat.get_left_right_filter()

    def cater(self, train_df: TrainDataframes):
        for feat in self.dc.train_features:
            categorized_feature = self.feat_cater(feat=feat, train_df=train_df, n=2)
            for n in range(3, self.settings.n_cat + 1):
                cat_feature = self.feat_cater(feat=feat, train_df=train_df, n=n)
                if cat_feature.calc_diff() > categorized_feature.calc_diff():
                    categorized_feature = cat_feature

            if not categorized_feature.is_diff_to_low():
                train_df.train_features.append(categorized_feature)

    def feat_cater(
        self, feat: Feature, train_df: TrainDataframes, n: int
    ) -> CategorizedFeature:
        match feat.statistical:
            case FeatureType.CATEGORIAL:
                return self.cat_cater_impl(feat=feat, train_df=train_df, n=n)
            case FeatureType.NUMERICAL:
                return self.cont_cater_impl(feat=feat, train_df=train_df, n=n)

        raise NotImplementedError(f"FeatureType {feat.statistical} not implemented")

    def cat_cater_impl(
        self, feat: Feature, train_df: TrainDataframes, n: int
    ) -> CategorizedFeature:
        df = (
            train_df.df_group.group_by(feat.name)
            .agg(pl.col("weight").sum(), pl.col("target_weight").sum())
            .sort(pl.col("weight"), descending=True)
        )
        default_cat = df.item(-1, feat.name)
        offstet = 0
        map = {}

        # First filter out large categories
        for i in range(n):
            if df.height == 0:
                break
            val = df.item(0, "weight")
            if val < 1.0 / (n - i):
                break
            map[df.item(0, feat.name)] = str(i)
            df = df.tail(-1).with_columns(pl.col("weight").truediv((1.0 - val)))
            offstet = i + 1

        if df.height == 0:
            return train_df.create_categorized_features(
                feat=feat,
                cuts=CatCats(
                    mapping=map, default=map.get(default_cat), feat_name=feat.name
                ),
            )
        # Sort by amount of target
        df = df.with_columns(
            pl.col("target_weight").truediv(pl.col("weight")).alias("target_rate")
        ).sort("target_rate", descending=False)

        # Build the column for the quatile
        df = df.with_columns(pl.col("weight").cum_sum().alias("cum_weight"))

        labels = [str(i) for i in range(offstet, n)]
        m = len(labels)
        breaks = [i / m for i in range(1, m)]

        # TODO the hole function is numerical instable. Write some test to find this error
        df = df.with_columns(
            pl.col("cum_weight")
            .cut(
                breaks=breaks,
                labels=labels,
                left_closed=True,
            )
            .alias("bin")
        )

        df_res = df.group_by("bin").agg(pl.col(feat.name))
        for row in df_res.to_dicts():
            map.update({feat: row.get("bin") for feat in row.get(feat.name)})
        return train_df.create_categorized_features(
            feat=feat,
            cuts=CatCats(
                mapping=map, default=map.get(default_cat), feat_name=feat.name
            ),
        )

    def cont_cater_impl(
        self, feat: Feature, train_df: TrainDataframes, n: int
    ) -> CategorizedFeature:
        df_sorted = train_df.df_group[[feat.name, "weight"]].sort(feat.name)
        cumulative_weights = df_sorted["weight"].cum_sum()

        res = []
        label = ["0"]
        idx = 0
        for i in range(1, n):
            quantile = i / n
            index = (cumulative_weights >= quantile).arg_true()[0]

            if index > 0:
                weight_below = cumulative_weights[index - 1]
                value_below = df_sorted[feat.name][index - 1]
                value_at_index = df_sorted[feat.name][index]
                fraction = (quantile - weight_below) / (
                    cumulative_weights[index] - weight_below
                )
                value = value_below + fraction * (value_at_index - value_below)
            else:
                value = df_sorted[feat.name][0]
            if value in res:
                continue
            idx += 1
            label.append(f"{idx}")
            res.append(value)

        return train_df.create_categorized_features(
            feat=feat,
            cuts=ContCats(cuts=res, labels=label, feat_name=feat.name),
        )



import json
import logging
from pathlib import Path

from pydantic import BaseModel, Field
import yaml

from pilz.model.pilz import Pilz, Pilze

logger = logging.getLogger(__name__)


class EvalSettings(BaseModel):
    in_folders: list[Path] = Field(description="Name where the pilze was stored")
    out_folder: Path = Field(description="Name where the evaluation is stored")
    max_parallel_where: int = Field(
        description="Number of points for roc curve", default=100
    )


class TrainSettings(BaseModel):
    n: int = Field(description="Number of Pilze per target value", default=1)
    out_folder: Path = Field(description="Name where the pilz will be stored")
    n_cat: int = Field(description="Number of categories in one featerue", default=3)
    max_depth: int = Field(description="Max depth of the pilz", default=10)

    max_eval_fit: int = Field(
        description="Wieviel sollen max für training gleichzeitig benutzt werden",
        default=1000,
    )
    min_eval_fit: int = Field(
        description="Wieviel müssen fürs training mindestens benutzt werden", default=10
    )
    n_dims: int = Field(
        description="Wieviele dimensionen sollen auf einmal benutzt werden", default=3
    )
    calcs_per_dim: int | None = Field(
        description="Wieviele berechnungen sollen pro dimension gemacht werden",
        default=5000,
    )
    frac_eval_cat: float = Field(
        description="Wie groß ist der anteil der für eval benutzt werden soll",
        default=0.5,
    )

    neutral_faktor: float = Field(
        description="Faktor wie groß der Unterschied für Neutral sein soll. Wert kleiner als 0 beduet kein Neutral state",
        default=-1.0,
    )

    def pilz_exists(self, target: str | int, n: int) -> bool:
        if self.out_folder.exists():
            if (self.out_folder / f"{target}" / f"{n}.json").exists():
                return True
        return False

    def save_pilz(self, pilz: Pilz, n: int):
        target = pilz.target
        out_folder = self.out_folder / f"{target}"
        logger.info(f"save pilz {n} for target {target} to {out_folder}")
        out_folder.mkdir(parents=True, exist_ok=True)
        out_file = out_folder / f"{n}.json"
        with out_file.open("w") as f:
            f.write(json.dumps(pilz.model_dump(), ensure_ascii=False, indent=2))



import logging
from pathlib import Path
from typing import Annotated
import typer
import yaml

from pilz.model.datacard import DataCard
from pilz.model.settings import EvalSettings, TrainSettings
from pilz.service.datacardhelper import dataCardHelper
from pilz.service.eval import Eval
from pilz.service.train import Train


logger = logging.getLogger(__name__)

app = typer.Typer()


# def main(config: Annotated[typer.FileText, typer.Option()]):
@app.command()
def train(
    datacard: Annotated[typer.FileText, typer.Option()],
    trainsettings: Annotated[typer.FileText, typer.Option()],
):
    train = Train(
        dc=DataCard(**yaml.safe_load(datacard)),
        settings=TrainSettings(**yaml.safe_load(trainsettings)),
    )
    train.run()


@app.command()
def eval(
    datacard: Annotated[typer.FileText, typer.Option()],
    evalsettings: Annotated[typer.FileText, typer.Option()],
):
    eval = Eval(
        dc=DataCard(**yaml.safe_load(datacard)),
        settings=EvalSettings(**yaml.safe_load(evalsettings)),
    )
    eval.run()


@app.command()
def infer(
    datacard: Annotated[typer.FileText, typer.Option()],
    evalsettings: Annotated[typer.FileText, typer.Option()],
):
    infer = Eval(
        dc=DataCard(**yaml.safe_load(datacard)),
        settings=EvalSettings(**yaml.safe_load(evalsettings)),
    )
    infer.run()


@app.command()
def create_dc(
    src: Annotated[Path, typer.Option()],
    out: Annotated[Path, typer.Option(prompt=True)],
):
    helper = dataCardHelper(src=src, out=out)
    helper.run()


# if __name__ == "__main__":
#     # Profilierung des Typer-Befehls
#     profiler = cProfile.Profile()
#     run_path = Path(
#         "/Users/swayand/PycharmProjects/pilz/examples/diabetes_prediction_challenge"
#     )
#     datacard = run_path / "diabetes_prediction_challenge.yaml"
#     trainsettings = run_path / "train_settings.yaml"
#     profiler.enable()

#     train(datacard=datacard.open(), trainsettings=trainsettings.open())
#     profiler.disable()
#     # sprofiler.print_stats(sort="time")

#     stats = pstats.Stats(profiler)
#     # Sort by different metrics
#     stats.sort_stats("cumulative").print_stats(
#         10
#     )  # Top 10 functions by cumulative time
#     # stats.sort_stats('calls').print_stats(10)       # Top 10 functions by call count
#     stats.sort_stats("time").print_stats(10)  # Top 10 functions by time



from enum import StrEnum
import logging
from typing import Literal
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class FeatureType(StrEnum):
    CATEGORIAL = "categorial"
    NUMERICAL = "numerical"


class Feature(BaseModel):
    name: str
    statistical: FeatureType
    type: str
    missing_value: int | str | float | None = None


class Target(BaseModel):
    feature_name: str
    values: list[int | str]


class DataCard(BaseModel):
    features: list[Feature]
    infos: dict
    target: Target
    test_files: list[str]
    train_files: list[str]

    @property
    def feature_names_sql_save(self) -> list[str]:
        return [f'"{feat.name}"' for feat in self.features]

    @property
    def feature_names(self) -> list[str]:
        return [feat.name for feat in self.features]

    @property
    def train_features(self) -> list[Feature]:
        return [feat for feat in self.features if feat.name != self.target.feature_name]



from pathlib import Path
from pydantic import BaseModel
import polars as pl
import typer
import yaml


class dataCardHelper(BaseModel):
    src: Path
    out: Path

    def run(self):
        df = self.get_df()

        while True:
            print(f"One of {df.columns}")
            target_feature = typer.prompt("Chose target column")
            if target_feature in df.columns:
                break
        target = {
            "feature_name": target_feature,
            "values": df[target_feature].unique().to_list(),
        }

        features = []
        for col in df.columns:
            akt_type = self.get_type(str(df[col].dtype))
            stat_type = "numerical" if akt_type in ["int", "float"] else "categorial"
            stat_type = "categorial" if target_feature == col else stat_type
            feature = {"name": col, "statistical": stat_type, "type": akt_type}
            if df[col].has_nulls():
                null_value = typer.prompt(
                    f"{col} has null values. What should be the null value (type {akt_type})?"
                )
                feature["missing_value"] = null_value

            features.append(feature)

        infos = typer.prompt("Infos about the dataset")
        dc = {
            "features": features,
            "target": target,
            "infos": {"bla": infos},
            "train_files": [str(self.src)],
            "test_files": [str(self.src)],
        }

        with self.out.open("w") as out_file:
            yaml.dump(dc, out_file)

    def get_type(self, dtype: str) -> str:
        match dtype:
            case "Int64":
                return "int"
            case "Float64":
                return "float"
            case "String":
                return "string"

        raise ValueError(f"Unsupported data type: {dtype}")

    def get_df(self) -> pl.DataFrame:
        if self.src.name.endswith(".csv"):
            return pl.read_csv(self.src)
        elif self.src.name.endswith(".parquet"):
            return pl.read_parquet(self.src)
        else:
            raise ValueError("Unsupported file type")



from itertools import batched
import logging
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class Spore(BaseModel):
    cut: list[str]
    score: float
    depth: str


class Pilz(BaseModel):
    spore: list[Spore]
    target: str | int

    def get_sql(self, res_name: str) -> str:
        return "CASE \n" + "\n".join(self.get_where_sql()) + f"END AS {res_name}"

    def get_split_sql(self, max_parallel_where: int) -> list[str]:
        where_cases = self.get_where_sql()
        return [
            "CASE \n" + "\n".join(batch) + f"\nELSE CAST(0 AS DOUBLE)\n END AS res_{i}"
            for i, batch in enumerate(batched(where_cases, max_parallel_where))
        ]

    def get_where_sql(self) -> list[str]:
        return [
            f"  WHEN {' AND '.join(spore.cut)} THEN CAST({spore.score} AS DOUBLE) "
            for spore in self.spore
        ]


class Pilze(BaseModel):
    pilze: dict[str, Pilz]
    target: str

    def get_sql(self) -> dict[str, str]:
        return {
            f"res_{name}": pilz.get_sql(f"res_{name}")
            for name, pilz in self.pilze.items()
        }



import itertools
import logging
from typing import Any
import mi_amore
import polars as pl
from pydantic import BaseModel
from dataclasses import dataclass

from sympy import And, Not, Or, SOPform, Symbol, simplify
from sympy.logic import POSform
from mi_amore.minimize import minimize
from sympy.logic.boolalg import BooleanTrue

from pilz.model.datacard import Feature
from pilz.model.filter import Filter, SympyContainer, SympyToSqlHelper
import sys

logger = logging.getLogger(__name__)


class CatsMixin(BaseModel):
    # labels: list[str]

    def get_bits(self, label: str) -> list[int]:
        bits = [0] * (len(self.labels))
        bits[int(label)] = 1
        return bits

    def get_bits_list(self, labels: list[str]) -> list[str]:
        return [self.get_bits(label=label) for label in labels]

    def get_sympy_from_bits(self, bits: list[int]):
        bits_sum = sum(bits)
        if bits_sum == len(bits):
            return simplify(True)
        if bits_sum == 1:
            for i, bit in enumerate(bits):
                if bit == 1:
                    return self.get_single_filter(i).get_sympy()
        if bits_sum == len(bits) - 1:
            for i, bit in enumerate(bits):
                if bit == 0:
                    return Not(self.get_single_filter(i).get_sympy())
        res = Or(
            *[
                self.get_single_filter(i).get_sympy()
                for i, bit in enumerate(bits)
                if bit == 1
            ]
        )
        # return res
        return simplify(res)


class CatCats(CatsMixin):
    mapping: dict[str | int, str]
    default: str
    feat_name: str

    @property
    def labels(self) -> list[str]:
        return list(set(self.mapping.values()))

    @property
    def inverted_mapping(self) -> dict[str, list[str | int]]:
        ret_map = {}
        for k, v in self.mapping.items():
            ret_map.setdefault(v, []).append(k)
        return ret_map

    @property
    def get_default_cuts(self) -> list[str]:
        defalut_cuts = []
        for i, v in self.inverted_mapping.items():
            if self.default != i:
                defalut_cuts += v
        return defalut_cuts

    def cut(self, df: pl.DataFrame) -> pl.Series:
        return df[self.feat_name].replace_strict(self.mapping, default=self.default)

    def get_single_filter(self, i: int) -> SympyContainer:
        i = f"{i}"
        if self.default == i:
            dev_cats = self.get_default_cuts
            if len(dev_cats) == 1:
                return SympyContainer(
                    feat_name=self.feat_name,
                    operator="!=",
                    value=dev_cats[0],
                    assumptions={},
                )
            return SympyContainer(
                feat_name=self.feat_name,
                operator="not in",
                value=dev_cats,
                assumptions={},
            )
        cats = self.inverted_mapping.get(i)

        if len(cats) == 1:
            return SympyContainer(
                feat_name=self.feat_name, operator="=", value=cats[0], assumptions={}
            )
        return SympyContainer(
            feat_name=self.feat_name, operator="in", value=cats, assumptions={}
        )


class ContCats(CatsMixin):
    cuts: list[float]
    labels: list[str]
    feat_name: str

    def cut(self, df: pl.DataFrame) -> pl.Series:
        # return series.cut(self.cuts)
        return df[self.feat_name].cut(self.cuts, labels=self.labels)

    def get_single_filter(self, i: int) -> SympyContainer:
        if i == 2 and len(self.cuts) == 1:
            logger.error("to low cuts")
            sys.exit()

        if i == 0:
            return SympyContainer(
                feat_name=self.feat_name,
                operator="<=",
                value=self.cuts[i],
                assumptions={"float": True},
            )

        if i == len(self.labels) - 1:
            return SympyContainer(
                feat_name=self.feat_name,
                operator=">",
                value=self.cuts[i - 1],
                assumptions={"float": True},
            )
        return SympyContainer(
            feat_name=self.feat_name,
            operator="between",
            value=[self.cuts[i - 1], self.cuts[i]],
            assumptions={"float": True},
        )


class CategorizedFeatureMixin:
    diff_df: pl.DataFrame
    cut_list: list[ContCats | CatCats]
    feature_list: list[str]
    neutral_faktor: float

    def calc_diff(
        self,
    ) -> float:
        # self.diff_sr.entropy()

        # return self.diff_df["diff"].abs().sqrt().sum()
        # return self.diff_df["diff"].abs().sum()
        return self.diff_df.with_columns(
            (pl.col("diff").abs() * pl.col("diff").abs().sqrt())
        )["diff"].sum()
        # return self.diff_df.with_columns((pl.col("diff") ** 2 ))['diff'].sum()

    def set_diff_df(
        self,
        df_count_target: pl.DataFrame,
        df_count_non_target: pl.DataFrame,
        join_on: list[str] | str,
    ):
        self.diff_df = (
            df_count_non_target.join(df_count_target, on=join_on, how="outer")
            .fill_null(0)
            .with_columns(
                (pl.col("proportion_right") - pl.col("proportion")).alias("diff"),
                pl.max_horizontal(
                    pl.col("proportion"), pl.col("proportion_right")
                ).alias("max_proportion"),
            )
        )

    def get_left_right_filter(self) -> tuple[Filter, Filter | None, Filter]:
        sorted_cut_list = sorted(self.cut_list, key=lambda el: len(el.labels))
        n_binary = 0
        mvars = []
        for el in sorted_cut_list:
            if len(el.labels) == 2:
                n_binary += 1
            else:
                mvars.append(len(el.labels))

        res_df = self.diff_df.with_columns(
            [
                pl.coalesce(
                    pl.col(cut.feat_name), pl.col(f"{cut.feat_name}_right")
                ).alias(f"{cut.feat_name}")
                for cut in sorted_cut_list
            ]
        )
        neutral_range = max(self.neutral_faktor, 0)
        target_lists = [
            res_df.filter(pl.col("diff") > neutral_range)[cut.feat_name].to_list()
            for cut in sorted_cut_list
        ]
        target_list_tupels = [akt_tuple for akt_tuple in zip(*target_lists)]

        non_target_lists = [
            res_df.filter(pl.col("diff") <= -neutral_range)[cut.feat_name].to_list()
            for cut in sorted_cut_list
        ]
        non_target_list_tupels = [akt_tuple for akt_tuple in zip(*non_target_lists)]

        target_cubes = []
        neutral_cubes = []
        non_target_cubes = []

        possible_labels = [cut.labels for cut in sorted_cut_list]
        for possible_combs in itertools.product(*possible_labels):
            bits = [
                flat_bit
                for label, cut in zip(possible_combs, sorted_cut_list)
                for flat_bit in cut.get_bits(label)
            ]
            if possible_combs in target_list_tupels:
                target_cubes.append(bits)
            elif possible_combs in non_target_list_tupels or self.neutral_faktor < 0:
                non_target_cubes.append(bits)
            else:
                neutral_cubes.append(bits)

        min_target_cubes = minimize(
            n_binary=n_binary,
            mvars=mvars,
            cubes_on=target_cubes,
        )
        min_neutral_cubes = minimize(
            n_binary=n_binary,
            mvars=mvars,
            cubes_on=neutral_cubes,
        )

        min_non_target_cubes = minimize(
            n_binary=n_binary,
            mvars=mvars,
            cubes_on=non_target_cubes,
        )
        # start at 0 next are binarys (two bits) and then the mvars
        low_high_cuts = list(itertools.accumulate([0] + [2] * n_binary + mvars))

        return (
            self.get_sympy_combine_filter(
                cut_list=sorted_cut_list,
                low_high_cuts=low_high_cuts,
                min_cubes=min_non_target_cubes,
            ),
            self.get_sympy_combine_filter(
                cut_list=sorted_cut_list,
                low_high_cuts=low_high_cuts,
                min_cubes=min_neutral_cubes,
            ),
            self.get_sympy_combine_filter(
                cut_list=sorted_cut_list,
                low_high_cuts=low_high_cuts,
                min_cubes=min_target_cubes,
            ),
        )

    @staticmethod
    def get_sympy_combine_filter(
        cut_list: list[ContCats | CatCats],
        low_high_cuts: list[int],
        min_cubes: list[list[int]],
    ) -> Filter | None:
        if len(min_cubes) == 0:
            return None
        return Filter(
            combine=Or(
                *[
                    And(
                        *[
                            cut.get_sympy_from_bits(cube[lh[0] : lh[1]])
                            for cut, lh in zip(
                                cut_list, itertools.pairwise(low_high_cuts)
                            )
                        ]
                    )
                    for cube in min_cubes
                ]
            )
        )


@dataclass
class CategorizedFeature(CategorizedFeatureMixin):
    feature: Feature
    cuts: ContCats | CatCats
    target_sr: pl.Series
    non_target_sr: pl.Series
    neutral_faktor: float

    def __post_init__(self):
        df_count_target = self.target_sr.value_counts()
        df_count_target = df_count_target.with_columns(
            (pl.col("count") / pl.col("count").sum()).alias("proportion")
        )

        df_count_non_target = self.non_target_sr.value_counts()
        df_count_non_target = df_count_non_target.with_columns(
            (pl.col("count") / pl.col("count").sum()).alias("proportion")
        )

        self.set_diff_df(
            df_count_target=df_count_target,
            df_count_non_target=df_count_non_target,
            join_on=self.feature.name,
        )

    @property
    def cut_list(self) -> list[ContCats]:
        return [self.cuts]

    def is_diff_to_low(self, threshold: float = 0.90) -> bool:
        max_wert = self.diff_df["max_proportion"].max()
        min_prop_of_max = self.diff_df.filter(pl.col("max_proportion") == max_wert)[
            "proportion", "proportion_right"
        ].min_horizontal()[0]
        return min_prop_of_max > threshold


class CombinedCategorizedFeature(CategorizedFeatureMixin):
    def __init__(
        self,
        train_features: tuple[CategorizedFeature],
        non_target_size: int,
        target_size: int,
        neutral_faktor: float,
    ):
        groub_by = [train.feature.name for train in train_features]
        non_target_df = pl.DataFrame([train.non_target_sr for train in train_features])
        self.neutral_faktor = neutral_faktor

        df_count_non_target = (
            non_target_df.group_by(groub_by)
            .len(name="proportion")
            .with_columns((pl.col("proportion") / non_target_size))
        )
        target_df = pl.DataFrame([train.target_sr for train in train_features])
        df_count_target = (
            target_df.group_by(groub_by)
            .len(name="proportion")
            .with_columns((pl.col("proportion") / target_size))
        )
        self.set_diff_df(
            df_count_target=df_count_target,
            df_count_non_target=df_count_non_target,
            join_on=groub_by,
        )
        self.cut_list = [train.cuts for train in train_features]


class TrainDataframes:
    def __init__(
        self,
        target_df: pl.DataFrame,
        non_target_df: pl.DataFrame,
        frac_eval_cat: float,
        min_size: int,
        neutral_faktor: float,
    ):
        self.target_df_size = target_df.height
        self.non_target_df_size = non_target_df.height

        self.min_size = min_size
        self.neutral_faktor = neutral_faktor

        self.n_count_target, n_group_target = self._calc_split(
            target_df.height, frac_eval_cat
        )
        self.target_df_count = target_df.head(self.n_count_target)
        target_df_group = target_df.tail(-self.n_count_target)
        target_df_group = target_df_group.with_columns(
            pl.lit(0.5 / max(1, n_group_target)).alias("weight")
        )
        target_df_group = target_df_group.with_columns(
            (2 * pl.col("weight")).alias("target_weight")
        )

        self.n_count_non_target, n_group_non_target = self._calc_split(
            non_target_df.height, frac_eval_cat
        )
        self.non_target_df_count = non_target_df.head(self.n_count_non_target)
        non_target_df_group = non_target_df.tail(-self.n_count_non_target)
        non_target_df_group = non_target_df_group.with_columns(
            pl.lit(0.5 / max(1, n_group_non_target)).alias("weight")
        )
        non_target_df_group = non_target_df_group.with_columns(
            pl.lit(0.0).alias("target_weight")
        )

        self.df_group = pl.concat([target_df_group, non_target_df_group])
        self.train_features: list[CategorizedFeature] = []

    def create_categorized_features(
        self, feat: Feature, cuts: ContCats | CatCats
    ) -> CategorizedFeature:
        target_sr = cuts.cut(df=self.target_df_count)
        non_target_sr = cuts.cut(df=self.non_target_df_count)
        return CategorizedFeature(
            feature=feat,
            cuts=cuts,
            target_sr=target_sr,
            non_target_sr=non_target_sr,
            neutral_faktor=self.neutral_faktor,
        )

    def _calc_split(self, n: int, frac: float):
        split = round(n * frac)
        if split < 1:
            return 1, 0
        if split >= n:
            return n - 1, 1
        return split, n - split

    def score(self) -> float:
        if self.non_target_df_size + self.target_df_size == 0:
            return 0.0
        return (self.target_df_size - self.non_target_df_size) / (
            self.non_target_df_size + self.target_df_size
        )

    def is_final_size(self) -> bool:
        return (
            self.target_df_size < self.min_size
            or self.non_target_df_size < self.min_size
        )


class EvalDataframe:
    def __init__(self, df: pl.DataFrame, target: str | int) -> None:
        self.df = df



import logging.config
import yaml

try:
    with open("logging.yaml", "r") as f:
        config = yaml.safe_load(f)
except FileNotFoundError:
    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "simple": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"}
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "simple",
                "level": "INFO",
            }
        },
        "root": {"handlers": ["console"], "level": "INFO"},
    }

logging.config.dictConfig(config)



import json
import os
from pathlib import Path
import sys
from typing import Optional
import duckdb
from pydantic import BaseModel
import yaml
import altair as alt

from pilz.model.pilz import Pilz, Pilze
from pilz.model.datacard import DataCard
from pilz.model.settings import EvalSettings
from pilz.service.darkwing import Darkwing
import polars as pl
import logging

logger = logging.getLogger(__name__)


class Eval(BaseModel):
    dc: DataCard
    settings: EvalSettings
    darkwing: Optional[Darkwing] = None

    def model_post_init(self, __context) -> None:
        self.darkwing = Darkwing(dc=self.dc)
        os.makedirs(self.settings.out_folder, exist_ok=True)
        super().model_post_init(__context)

    def run(self):
        logger.info("Start Eval")

        df_res = self.darkwing._get_pl_eval_df(col_sql=self.dc.target.feature_name)

        roc_plots = None

        for target in self.dc.target.values:
            logger.info(f"Evaluate target {target}")
            target_pilze = self.load_pilze(target=f"{target}")

            sr = self.darkwing.get_eval_sr(
                pilze=target_pilze,
                max_parallel_where=self.settings.max_parallel_where,
            )
            df_res = df_res.with_columns(sr)
            hist_plot = self.plot_hist(df=df_res, target_label=target, stack=None)
            hist_plot.save(self.settings.out_folder / f"{target}_hist.html")

            rel_pdf_plot = self.plot_hist(
                df=df_res, target_label=target, stack="normalize"
            )
            rel_pdf_plot.save(self.settings.out_folder / f"{target}_pdf_rel.html")

            new_roc_plot = self.plot_roc(df=df_res, target_label=target)
            new_roc_plot.save(self.settings.out_folder / f"{target}_roc.html")
            if roc_plots is None:
                roc_plots = new_roc_plot
            else:
                roc_plots = roc_plots + new_roc_plot

        roc_plots.save(self.settings.out_folder / "all_roc.html")

        if len(self.dc.target.values) > 1:
            df_multi_res = self.add_multi_class_result(
                df=df_res, target_values=self.dc.target.values
            )
            df_res = pl.concat(
                [df_res, df_multi_res],
                how="horizontal",
            )

            multi_res_plot = (
                df_res.group_by(self.dc.target.feature_name)
                .agg(pl.col("correct").sum() / pl.count())
                .sort(self.dc.target.feature_name)
                .plot.bar(x=self.dc.target.feature_name, y="correct")
            )
            prec = df_res["correct"].sum() / df_res["correct"].count()
            all_pred_labels = (
                alt.Chart(pl.DataFrame({"correct": [prec]}))
                .mark_rule()
                .encode(y="correct")
            )
            (multi_res_plot + all_pred_labels).save(
                self.settings.out_folder / "multi_class_result.html"
            )
        logger.info(f"{df_res}")

    def add_multi_class_result(self, df: pl.DataFrame, target_values: list[str | int]):
        greates = ", ".join([f'"{value}"' for value in target_values])
        cases = [
            f'WHEN "{value}" = GREATEST({greates}) THEN {value}'
            for value in target_values
        ]
        p_name = f"predicted_{self.dc.target.feature_name}"
        return (
            duckdb.sql(f"""
            SELECT
                {self.dc.target.feature_name},
                CASE
                    {" \n ".join(cases)}
                END AS {p_name}
            FROM df
        """)
            .pl()
            .with_columns(
                (pl.col(f"{self.dc.target.feature_name}") == pl.col(p_name))
                .cast(pl.Int64)
                .alias("correct")
            )
            .select(p_name, "correct")
        )

    def load_pilze(self, target: str) -> Pilze:
        data = {}
        for folder in self.settings.in_folders:
            target_folder = folder / target
            if not target_folder.exists():
                logger.error(f"folder {target_folder} does not exist. ")
                break
            for file in target_folder.glob("*.json"):
                f_name = str(file).rstrip(".json").replace("/", "_")
                with file.open("rb") as f:
                    logger.info(f"Load Pilz from {file}")
                    data[f_name] = Pilz.model_validate_json(f.read())
        return Pilze(pilze=data, target=target)

    def plot_hist(
        self, df: pl.DataFrame, target_label: str | int, stack: bool | str | None
    ):
        return (
            df.with_columns(
                pl.col(self.dc.target.feature_name).eq(target_label).alias("Result")
            )
            .plot.bar(
                x=alt.X(f"{target_label}", bin=alt.Bin(maxbins=40)),
                y=alt.Y("count()", stack=stack),
                color="Result",
            )
            .properties(title="Histogramm der Werte")
        ).configure_mark(
            opacity=0.7  # Transparenz einstellen, kann angepasst werden
        )

    def plot_roc(self, df: pl.DataFrame, target_label: str | int):
        score_feat = f"{target_label}"

        df_plot = duckdb.sql(f"""
            SELECT fp,tp
            FROM (
                SELECT
                    ROUND(SUM(CASE WHEN {self.dc.target.feature_name} == '{target_label}' THEN 1 ELSE 0 END) OVER (ORDER BY "{score_feat}" DESC) / SUM(CASE WHEN {self.dc.target.feature_name} == '{target_label}' THEN 1 ELSE 0 END) OVER (), 4)  AS tp,
                    ROUND(SUM(CASE WHEN {self.dc.target.feature_name} == '{target_label}' THEN 0 ELSE 1 END) OVER (ORDER BY "{score_feat}" DESC) / SUM(CASE WHEN {self.dc.target.feature_name} == '{target_label}' THEN 0 ELSE 1 END) OVER (), 4)  AS fp
                FROM df
                )
            GROUP BY fp,tp
            ORDER BY fp,tp
        """).pl(
            # ).vstack(pl.DataFrame({'fp': 0.0, 'tp': 0.0})
        )
        df_shift = df_plot.with_columns(
            [
                pl.col("fp").shift(-1).alias("fp_next"),
                pl.col("tp").shift(-1).alias("tp_next"),
            ]
        )
        auc = (
            df_shift.with_columns(
                (
                    (pl.col("fp_next") - pl.col("fp"))
                    * (pl.col("tp") + pl.col("tp_next"))
                    / 2
                ).alias("trap")
            )
            .select(pl.col("trap").sum())
            .item()
        )

        plot = (
            df_plot.with_columns(
                pl.lit(f"{score_feat} (AUC {auc:.4f})").alias("target")
            )
            .plot.line(x="fp", y="tp", color="target")
            .interactive()
        )

        return plot



from itertools import batched
import logging
from pathlib import Path
import sys
from typing import Any
import duckdb
from pydantic import BaseModel
import polars as pl
from sympy import And, Not
import yaml

from pilz.model.pilz import Pilze
from pilz.model.datacard import DataCard
from pilz.model.dataframes import EvalDataframe, TrainDataframes
from pilz.model.filter import Filter, SympyToSqlHelper
from pilz.model.settings import TrainSettings
from sympy.logic.boolalg import Boolean


logger = logging.getLogger(__name__)


class Darkwing(BaseModel):
    dc: DataCard
    df_train_cach: Any | None = None
    df_test_cache: Any | None = None

    def read_akt_train(
        self,
        targer_filter: Filter,
        train_settings: TrainSettings,
        akt_filters: list[Filter] = [],
    ) -> TrainDataframes:
        """
        Reads a list of CSV files into a DuckDB relation.

        Args:
            filepaths: A list of filepaths to CSV files.

        Returns:
            A DuckDB relation containing the data from all CSV files.
        """

        full_filters_target_list = [targer_filter.combine] + [
            f.combine for f in akt_filters
        ]
        full_filter_target = (
            And(*full_filters_target_list)
            if len(full_filters_target_list) > 1
            else full_filters_target_list[0]
        )

        target_df = self._get_pl_train_df(
            full_filters=full_filter_target,
            max_eval_fit=train_settings.max_eval_fit,
        )

        full_filters_non_target_list = [Not(targer_filter.combine)] + [
            f.combine for f in akt_filters
        ]
        full_filters_non_target = (
            And(*full_filters_non_target_list)
            if len(full_filters_non_target_list) > 1
            else full_filters_non_target_list[0]
        )

        non_target_df = self._get_pl_train_df(
            full_filters=full_filters_non_target,
            max_eval_fit=train_settings.max_eval_fit,
        )

        return TrainDataframes(
            target_df=target_df,
            non_target_df=non_target_df,
            frac_eval_cat=train_settings.frac_eval_cat,
            min_size=train_settings.min_eval_fit,
            neutral_faktor=train_settings.neutral_faktor,
        )

    def _get_pl_train_df(self, full_filters: Boolean, max_eval_fit) -> pl.DataFrame:
        df = self.get_cached_train_df()
        sql_str = SympyToSqlHelper.to_sql_where(full_filters)

        sql = f"""
                SELECT {", ".join(self.dc.feature_names_sql_save)}
                FROM df
                WHERE {sql_str}
                ORDER BY RANDOM()
                LIMIT {max_eval_fit};
              """

        return duckdb.sql(sql).pl().sample(fraction=1, shuffle=True)
    

    def get_df_from_files(self, files: list[str]) -> pl.DataFrame:
        if files[0].endswith(".csv"):
                df = pl.read_csv(",".join(files))
        elif files[0].endswith(".parquet"):
                df = pl.read_parquet(",".join(files))
        else:
                raise ValueError("Unsupported file type")
        for feat in self.dc.features:
                if feat.missing_value is not None:
                    # Todo move this to sql statment
                    df = df.with_columns(
                        pl.col(feat.name).fill_null(feat.missing_value)
                    )
        return df

    def get_cached_train_df(self) -> pl.DataFrame:
        if self.df_train_cach is None:
             self.df_train_cach = self.get_df_from_files(self.dc.train_files)

        return self.df_train_cach

    def get_eval_sr(
        self, pilze: Pilze, max_parallel_where: int
    ) -> EvalDataframe:
        df_sum = None
        for name, pilz in pilze.pilze.items():
            if len(pilz.spore) < max_parallel_where:
                case_sql = pilz.get_sql(name)
                df = self._get_pl_eval_df(col_sql=case_sql)
            else:
                where_cases = pilz.get_split_sql(
                    max_parallel_where=max_parallel_where
                )
                df = (
                    pl.concat(
                        [
                            self._get_pl_eval_df(col_sql=sub_sql)
                            for sub_sql in where_cases
                        ],
                        how="horizontal",
                    )
                    .sum_horizontal()
                    .alias(name)
                    .to_frame()
                )

            df_sum = df_sum.hstack(df) if df_sum is not None else df

        # return df_sum.transpose().median().transpose().to_series().alias(pilz.target)

        return (df_sum.mean_horizontal()).alias(pilze.target)

        # for feat_name, case_sql in pilz.get_sql().items():
        #    df = self._get_pl_eval_df(case_sql=case_sql)
        #    logger.info(f"Evaluate {df}")
        #    logger.info(f"{case_sql}")

    def _get_pl_eval_df(self, col_sql: str) -> pl.DataFrame:
        df = self.get_cached_eval_df()
        sql = f"""
                SELECT {col_sql}\n
                FROM df; 
              """
        return duckdb.sql(sql).pl()

    def get_cached_eval_df(self) -> pl.DataFrame:
        if self.df_test_cache is None:
            self.df_test_cache = self.get_df_from_files(self.dc.test_files)
        return self.df_test_cache
